""" Exceptions raised by the module """
class BaseResponseException(Exception):
    def __new__(cls, *args, **kwargs):
        return super(BaseResponseException, cls).__new__(BaseResponseException)


class HTTPException(BaseResponseException):
    def __init__(self, code, message):
        super().__init__(f'Status {code}: {message}')

class GatewayConnectionRefused(BaseResponseException):
    """ Raised when connecting to gateway fails """


class CannotOverideTokenWarning(Warning):
    """ Warned when cannot use provided token due to binded token present """



# Basic heartbeat controller
import asyncio
from .signals import gateway


class KeepAlive(object):
    def __init__(self, identity, ws, helloPacket: dict):
        self._ws = ws
        self.packet = helloPacket
        self.identity = identity

    async def run(self):
        packet = self.packet

        await self._ws.send_json(self.identity)

        while True:
            if packet['op'] != gateway.HELLO:
                raise ValueError('Invalid hello packet provided')

            await asyncio.sleep((packet['d']['heartbeat_interval'] / 1000))

            await self._ws.send_json(await self.get_payload())

    async def get_payload(self):
        return {
            "op": gateway.HEARTBEAT,
            "d": gateway.SEQUENCE
        }



# A simple base client for handling responses from discord
import asyncio
import typing
import warnings
from zlib import compress
import acord
import sys
import traceback
from inspect import iscoroutinefunction

from acord.core.decoders import ETF, JSON, decompressResponse
from acord.core.signals import gateway

from .core.http import HTTPClient
from .errors import *
from functools import wraps

from typing import (
    Union, Callable
)

from acord.models import User


class Client(object):
    """
    Client for interacting with the discord API

    Parameters
    ----------
    loop: :class:`~asyncio.AbstractEventLoop`
        An existing loop to run the client off of
    token: :class:`str`
        Your API Token which can be generated at the developer portal
    tokenType: typing.Union[BEARER, BOT]
        The token type, which controls the payload data and restrictions.

        .. warning::
            If BEARER, do not use the `run` method. Your able to access data normally.
    commandHandler: :class:`~typing.Callable`
        An optional command handler, defaults to the built-in handler at :class:`~acord.DefaultCommandHandler`.
        
        **Parameters passed though:**

        * Message: :class:`~acord.Message`
        * UpdatedCache: :class:`bool`
    """
    def __init__(self, *,
        loop: asyncio.AbstractEventLoop = asyncio.get_event_loop(),
        token: str = None,
        encoding: str = "JSON",
        compress: bool = False,
        commandHandler: Callable = None,
    ) -> None:

        self.loop = loop
        self.token = token

        self._events = dict()
        self.commandHandler = commandHandler

        # Gateway connection stuff
        self.encoding = encoding
        self.compress = compress

        # Others
        self.session_id = None
        self.gateway_version = None
        self.user = None

    def bindToken(self, token: str) -> None:
        self._lruPermanent = token

    def event(self, func):
        if not iscoroutinefunction(func):
            raise ValueError('Provided function was not a coroutine')
        
        eventName = func.__qualname__
        if eventName in self._events:
            self._events[eventName].append(func)
        else:
            self._events.update({eventName: [func]})

        return func

    def on_error(self, event_method):
        acord.logger.error('Failed to run event "{}".'.format(event_method))

        print(f'Ignoring exception in {event_method}', file=sys.stderr)
        traceback.print_exc()

    async def dispatch(self, event_name: str, *args, **kwargs) -> None:
        if not event_name.startswith('on_'):
            event_name = 'on_' + event_name
        acord.logger.info('Dispatching event: {}'.format(event_name))

        events = self._events.get(event_name, [])
        acord.logger.info('Total of {} events found for {}'.format(len(events), event_name))
        for event in events:
            try:
                await event(*args, **kwargs)
            except Exception:
                self.on_error(event)

    async def handle_websocket(self, ws):
        async for message in ws:
            await self.dispatch('socket_recieve')

            data = message.data
            if type(data) is bytes:
                data = decompressResponse(data)
            
            if not data:
                continue

            if not data.startswith('{'):
                data = ETF(data)
            else:
                data = JSON(data)

            if data['op'] == gateway.INVALIDSESSION:
                acord.logger.error('Invalid Session - Reconnecting Shortly')
                raise GatewayConnectionRefused('Invalid session data, currently not handled in this version')

            if data['t'] == 'READY':
                await self.dispatch('ready')

                self.session_id = data['d']['session_id']
                self.gateway_version = data['d']['v']
                self.user = User(**data['d']['user'])

                continue

            if data['op'] == gateway.HEARTBEATACK:
                await self.dispatch('heartbeat')

    def resume(self):
        """ Resumes a closed gateway connection """

            

    def run(self, token: str = None, *, reconnect: bool = True):
        if (token or self.token) and getattr(self, '_lruPermanent', False):
            warnings.warn("Cannot use current token as another token was binded to the client", CannotOverideTokenWarning)
        token = getattr(self, '_lruPermanent', None) or (token or self.token)

        if not token:
            raise ValueError('No token provided')

        self.http = HTTPClient(loop=self.loop)
        self.token = token

        # Login to create session
        self.loop.run_until_complete(self.http.login(token=token))

        coro = self.http._connect(
            token,
            encoding=self.encoding, 
            compress=self.compress
        )

        # Connect to discord, send identity packet + start heartbeat
        ws = self.loop.run_until_complete(coro)
        
        self.loop.run_until_complete(self.dispatch('connect'))
        acord.logger.info('Connected to websocket')

        self.loop.run_until_complete(self.handle_websocket(ws))



import pydantic


class User(pydantic.BaseModel):
    verified: bool
    username: str
    mfa_enabled: bool
    id: int
    flags: int
    discriminator: int
    bot: str
    avatar: str

    email: str = None



"""
Main HTTP connection and websocket interaction between discord and your application
"""
try:
    import uvloop

    uvloop.install()
except ImportError:
    __import__('warnings').warn('Failed to import UVLoop, it is recommended to install this library\npip install uvloop', ImportWarning)

import asyncio
import typing
import aiohttp
import acord
import sys

from acord.errors import GatewayConnectionRefused, HTTPException
from . import helpers
from .heartbeat import KeepAlive
from .decoders import *
from .signals import gateway

class HTTPClient(object):
    """
    Base client used to connection and interact with the websocket.

    Parameters
    ----------
    loop: :class:`~asyncio.AbstractEventLoop`
        A pre-existing loop for aiohttp to run of, defaults to ``asyncio.get_event_loop()``
    reconnect: :class:`bool`
        Attempt to reconnect to gateway if failed, If set to a integer, it will re-attempt n times.
    wsTimeout: :class:`~aiohttp.ClientTimeout`
        Custom timeout configuration for 
    **payloadData: :class:`dict`
        A dictionary of payload data to be sent with any request
        
        .. note::
            This information can be overwritten with each response
    """
    def __init__(self,
        token: str = None,
        connecter: typing.Optional[aiohttp.BaseConnector] = None,
        wsTimeout: aiohttp.ClientTimeout = aiohttp.ClientTimeout(60, connect=None),
        proxy: typing.Optional[str] = None,
        proxy_auth: typing.Optional[aiohttp.BasicAuth] = None,
        loop: typing.Optional[asyncio.AbstractEventLoop] = asyncio.get_event_loop(),
        unsync_clock: bool = True,

    ) -> None:
        self.token = token
        self.loop = loop
        self.wsTimeout = wsTimeout
        self.connector = connecter
        
        self._ws_connected = False
        self.proxy = proxy
        self.proxy_auth = proxy_auth
        self.use_clock = not unsync_clock

        user_agent = "ACord - https://github.com/Mecha-Karen/ACord {0} Python{1[0]}.{1[1]} aiohttp/{2}"
        self.user_agent = user_agent.format(
            acord.__version__, sys.version, aiohttp.__version__
        )

    def getIdentityPacket(self, intents = 0): 
        return {
            "op": gateway.IDENTIFY,
            "d": {
                "token": self.token,
                "intents": 513,
                "properties": {
                    "$os": sys.platform,
                    "$browser": "acord",
                    "$device": "acord"
                }
            }
        }

    def updatePayloadData(self, overwrite: bool = False, **newData) -> None:
        if overwrite:
            self.startingPayloadData = newData
        else:
            self.startingPayloadData = {**self.startingPayloadData, **newData}

    async def login(self, *, token: str) -> None:
        """ Define a session for the http client to use. """
        self._session = aiohttp.ClientSession(connector=self.connector)
        ot = self.token

        self.token = token

        try:
            data = await self.request(
                helpers.Route("GET", path="/users/@me")
            )
        except HTTPException as exc:
            self.token = ot
            acord.logger.error('Failed to login to discord, improper token passed')
            raise GatewayConnectionRefused('Invalid or Improper token passed') from exc

        return data


    async def _fetchGatewayURL(self, token):
        uri = helpers.buildURL('gateway', 'bot')
        
        async with self._session.get(uri, headers={'Authorization': f"Bot {token}"}) as resp:
            data = await resp.json()

            return data

    async def _connect(self, token: str, *, 
        encoding: helpers.GATEWAY_ENCODING, compress: int = 0,
        **identityPacketKwargs
    ) -> None:
        if not getattr(self, '_session', False):
            acord.logger.warn('Session not defined, user not logged in. Called login manually')
            await self.login(token=(token or self.token))

        self.encoding = encoding
        self.compress = compress

        respData = await self._fetchGatewayURL(token)
        GATEWAY_WEBHOOK_URL = respData['url']

        GATEWAY_WEBHOOK_URL += f'?v={helpers.API_VERSION}'
        GATEWAY_WEBHOOK_URL += f'&encoding={encoding.lower()}'

        if compress:
            GATEWAY_WEBHOOK_URL += "&compress=zlib-stream"

        acord.logger.info('Generated websocket url: %s' % GATEWAY_WEBHOOK_URL)

        kwargs = {
            'proxy_auth': self.proxy_auth,
            'proxy': self.proxy,
            'max_msg_size': 0,
            'timeout': self.wsTimeout.total,
            'autoclose': False,
            'headers': {
                'User-Agent': self.user_agent,
            },
            'compress': compress
        }

        ws = await self._session.ws_connect(GATEWAY_WEBHOOK_URL, **kwargs)

        helloRecv = await ws.receive()
        data = helloRecv.data
        if compress:
            data = decompressResponse(data)
        if not data.startswith('{'):
            data = ETF(data)
        else:
            data = JSON(data)

        self._ws_connected = True
        self.ws = ws

        self.loop.create_task(KeepAlive(self.getIdentityPacket(**identityPacketKwargs), ws, data).run())

        return ws

    async def request(self, route: helpers.Route, data: dict = None, **payload) -> None:
        url = route.url

        headers = payload

        headers['Authorization'] = "Bot " + self.token
        headers['User-Agent'] = self.user_agent

        kwargs = dict()
        kwargs['data'] = data
        kwargs['headers'] = headers

        resp = await self._session.request(
            method=route.method,
            url=url,
            **kwargs
        )

        return resp
        
    @property
    def connected(self):
        return self._ws_connected



import zlib
import json

ZLIB_SUFFIX = b'\x00\x00\xff\xff'
INFLATOR = zlib.decompressobj()
BUFFER = bytearray()


def decompressResponse(msg):
    if type(msg) is bytes:
        BUFFER.extend(msg)

        if len(msg) < 4 or msg[-4:] != b'\x00\x00\xff\xff':
            return
        msg = INFLATOR.decompress(BUFFER)
        msg = msg.decode('utf-8')
        BUFFER.clear()

    return msg


def ETF(msg):
    raise NotImplementedError()


def JSON(msg):
    return json.loads(msg)



"""
ACord - An API wrapper for the discord API.

Created by Mecha Karen, and is licensed under the GNU GENERAL PUBLIC LICENSE.
"""
from typing import NamedTuple, Literal
import logging

from .client import Client


logger = logging.getLogger("ACord")
__file__ = __import__("os").path.abspath(__file__)
__doc__ = "An API wrapper for the discord API"
__version__ = "0.0.1a"
__author__ = "Mecha Karen"

class VersionInfo(NamedTuple):
    major: int
    minor: int
    micro: int
    level: Literal["Alpha", "Beta", "Stable", "Final"]


version_info: VersionInfo = VersionInfo(major=0, minor=0, micro=1, level="Pre-Alpha")



"""
All version related info for connecting to the gateway.
"""
import yarl
from typing import Optional, Literal

API_VERSION = 9
BASE_API_URL = "https://discord.com/api"
GATEWAY_ENCODING = Literal["JSON", "ETF"]


def buildURL(*paths, **parameters) -> str:
    URI = f'{BASE_API_URL}/v{API_VERSION}'
    for path in paths:
        URI += f'/{path}'

    if not parameters:
        return URI

    URI += '?'

    for key, value in parameters.items():
        URI += f'{key}={value}&'

    return yarl.URL(URI)


class Route(object):
    """ Simple object representing a route """
    def __init__(self, method: str = "GET", *paths, path="/", **parameters):
        if path:
            paths = path.split('/')
        
        self.paths = paths          # Building url with yarn
        self.path = '/'.join(paths)
        self.parameters = parameters
        self.method = method
        self.url = buildURL(*paths, **parameters)

        self.channel_id: Optional[...] = parameters.get('channel_id')
        self.guild_id: Optional[...] = parameters.get('guild_id')
        self.webhook_id: Optional[...] = parameters.get('webhook_id')
        self.webhook_token: Optional[str] = parameters.get('webhook_token')

    @property
    def bucket(self):
        return f'{self.channel_id}:{self.guild_id}:{self.path}'



DISPATCH        = 0
HEARTBEAT       = 1
IDENTIFY        = 2
PRESENCE        = 3
VOICE           = 4
RESUME          = 6
RECONNECT       = 7
GUILDMEMBERS    = 8
INVALIDSESSION  = 9
HELLO           = 10
HEARTBEATACK    = 11

# Error Codes
UNKNOWN             = 4000
UNKNOWN_OP          = 4001
DECODE_ERROR        = 4002
FORBIDDEN           = 4003
AUTH_FAILED         = 4004
AUTH_COMPLETED      = 4005
FAILED_SEQUENCE     = 4007
RATELIMIT           = 4008
SESSION_TIMED_OUT   = 4009
INVALID_SHARD       = 4010
SHARD_REQUIRED      = 4011
INVALID_GATEWAY_VER = 4012
INVALID_INTENTS     = 4013
DISALLOWED_INTENT   = 4014

SEQUENCE = None

def heartbeatPacket():
    return {
        "op": HEARTBEAT,
        "d": SEQUENCE
    }



from setuptools import setup
import re
import sys

versionInfo = sys.version_info

if (versionInfo.major < 3) and (versionInfo.minor < 8):
    sys.exit('Cannot install ACord on python version below 3.8, Please upgrade!')

version = ""

with open("acord/__init__.py") as f:
    contents = f.read()

    _match = re.search(
        r'^__version__\s*=\s*[\'"]([^\'"]*)[\'"]', contents, re.MULTILINE
    )

    version = _match.group(1)

if not version:
    raise RuntimeError("Cannot resolve version")


classifiers = [
    "Development Status :: 1 - Planning",
    "Intended Audience :: Developers",
    "Operating System :: OS Independent",
    "License :: OSI Approved :: GNU General Public License (GPL)",
    
    "Programming Language :: Python :: Implementation :: CPython",
    "Programming Language :: Python :: 3",
]

packages = [
    "acord",
    "acord.core",
    "acord.core.signals",
    "acord.models"
]

setup(
    name="ACord",
    version=version,
    description="An API wrapper for discord",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/Mecha-Karen/acord",
    project_urls={
        "Documentation": "https://docs.mechakaren.xyz/acord",
        "Issue tracker": "https://github.com/Mecha-Karen/acord/issues",
        "Organisation": "https://github.com/Mecha-Karen",
    },
    author="Mecha Karen",
    author_email="admin@mechakaren.xyz",
    license="GNU License",
    classifiers=classifiers,
    packages=packages,
    install_requires=["aiohttp", "pydantic"],
)
